package com.yihu.elasticsearch;

import com.alibaba.druid.sql.ast.SQLExpr;
import com.alibaba.druid.sql.ast.expr.SQLQueryExpr;
import com.alibaba.druid.sql.parser.SQLExprParser;
import com.alibaba.fastjson.JSONObject;
import com.yihu.elasticsearch.model.ESIDEntity;
import io.searchbox.client.JestClient;
import io.searchbox.client.JestResult;
import io.searchbox.core.*;
import org.elasticsearch.action.search.SearchResponse;
import org.elasticsearch.client.Client;
import org.nlpcn.es4sql.domain.Select;
import org.nlpcn.es4sql.jdbc.ObjectResult;
import org.nlpcn.es4sql.jdbc.ObjectResultsExtractor;
import org.nlpcn.es4sql.parse.ElasticSqlExprParser;
import org.nlpcn.es4sql.parse.SqlParser;
import org.nlpcn.es4sql.query.AggregationQueryAction;
import org.nlpcn.es4sql.query.DefaultQueryAction;
import org.nlpcn.es4sql.query.SqlElasticSearchRequestBuilder;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Component;

import java.io.IOException;
import java.text.DecimalFormat;
import java.util.*;


/**
 * Utils - 主要基于(jestClient)的搜索工具
 * Created by chenweida on 2017/6/2.
 */
@Component
public class ElasticSearchHelper {

    private Logger logger = LoggerFactory.getLogger(ElasticSearchHelper.class);

    @Autowired
    private ElasticSearchPool elasticSearchPool;
    @Autowired
    private JestClient jestClient;

    /**
     * 新增
     *
     * @param index
     * @param type
     * @param sources
     * @return
     */
    public <T> Boolean save (String index, String type, List<T> sources) throws IOException {
        Bulk.Builder bulk = new Bulk.Builder().defaultIndex(index).defaultType(type);
        sources.forEach(item -> {
            Index indexObj = new Index.Builder(item).build();
            bulk.addAction(indexObj);
        });
        BulkResult br = jestClient.execute(bulk.build());
        logger.debug("save flag: " + br.isSucceeded());
        return br.isSucceeded();
    }

    public Boolean save (String index, String type, String source) throws IOException {
        Bulk.Builder bulk = new Bulk.Builder().defaultIndex(index).defaultType(type);
        Index indexObj = new Index.Builder(source).build();
        bulk.addAction(indexObj);
        BulkResult br = jestClient.execute(bulk.build());
        logger.debug("save flag: " + br.isSucceeded());
        return br.isSucceeded();
    }

    /**
     * 自定义ID
     *
     * @param index
     * @param type
     * @param source
     * @param idFieldString
     * @return
     */
    public Boolean saveWithCustomId (String index, String type, String source, String idFieldString) throws IOException {
        Bulk.Builder bulk = new Bulk.Builder().defaultIndex(index).defaultType(type);
        JSONObject jsonObject = (JSONObject) (JSONObject.parse(source));
        Index indexObj = new Index.Builder(source).id(jsonObject.getString(idFieldString)).
                build();
        bulk.addAction(indexObj);
        BulkResult br = jestClient.execute(bulk.build());
        logger.debug("save flag: " + br.isSucceeded());
        return br.isSucceeded();
    }

    /**
     * 自定义ID
     *
     * @param index
     * @param type
     * @param sources
     * @param idFieldString
     * @return
     */
    public Boolean saveBulkWithCustomId (String index, String type, List<String> sources, String idFieldString) throws IOException {
        Bulk.Builder bulk = new Bulk.Builder().defaultIndex(index).defaultType(type);
        sources.forEach(item -> {
            JSONObject jsonObject = (JSONObject) (JSONObject.parse(item));
            Index indexObj = new Index.Builder(item).id(jsonObject.getString(idFieldString)).
                    build();
            bulk.addAction(indexObj);
        });
        BulkResult br = jestClient.execute(bulk.build());
        logger.debug("save flag: " + br.isSucceeded());
        return br.isSucceeded();
    }

    /**
     * 更新
     *
     * @param index
     * @param type
     * @param sources
     * @return
     */
    public <T> Boolean update (String index, String type, List<T> sources) throws IOException {
        Bulk.Builder bulk = new Bulk.Builder().defaultIndex(index).defaultType(type);
        sources.forEach(item -> {
            JSONObject jo = new JSONObject();
            jo.put("doc", item);
            Update indexObj = new Update.Builder(jo.toString()).index(index).type(type).id(((ESIDEntity) item).getId()).build();
            bulk.addAction(indexObj);
        });
        BulkResult br = jestClient.execute(bulk.build());
        logger.debug("update flag: " + br.isSucceeded());
        return br.isSucceeded();
    }

    /**
     * 更新
     * @param index
     * @param type
     * @param list
     * @return
     */
    public Boolean updateByMap(String index, String type, List<Map<String, Object>> list) throws IOException {
        Bulk.Builder bulk = (new Bulk.Builder()).defaultIndex(index).defaultType(type);
        Iterator var10 = list.iterator();
        while (var10.hasNext()) {
            Map map = (Map)var10.next();
            JSONObject jo = new JSONObject();
            jo.put("doc", map);
            Update indexObj = (((new Update.Builder(jo.toString())).index(index)).type(type)).id(String.valueOf(map.get("id"))).build();
            bulk.addAction(indexObj);
        }
        BulkResult br = jestClient.execute(bulk.build());
        logger.debug("update flag: " + br.isSucceeded());
        return br.isSucceeded();
    }

    /**
     * 更新
     *
     * @param index
     * @param type
     * @param _id
     * @param source
     * @return
     */
    public boolean update(String index, String type, String _id, JSONObject source) throws IOException {
        JSONObject docSource = new JSONObject();
        docSource.put("doc", source);
        Update update = new Update.Builder(docSource).index(index).type(type).id(_id).build();
        JestResult jestResult = jestClient.execute(update);
        logger.debug("update info: " + jestResult.isSucceeded());
        return true;
    }

    /**
     * 删除
     */
    public boolean delete(String index, String type, List<Map<String, Object>> datas) throws IOException {
        //根据id批量删除
        Bulk.Builder bulk = new Bulk.Builder().defaultIndex(index).defaultType(type);
        for (Map map : datas) {
            if (!map.containsKey("id") || !map.containsKey("_id")) {
                continue;
            }
            Delete indexObj = null;
            if (null != map.get("_id")) {
                indexObj = new Delete.Builder(map.get("_id").toString()).build();
            } else if (null != map.get("id")) {
                indexObj = new Delete.Builder(map.get("id").toString()).build();
            }
            bulk.addAction(indexObj);
        }
        BulkResult br = jestClient.execute(bulk.build());
        logger.debug("delete data count: " + datas.size());
        logger.debug("delete flag: " + br.isSucceeded());
        return br.isSucceeded();
    }

    public SearchResult search(String index, String type, String queryStr) throws IOException {
        Search search = ((new Search.Builder(queryStr)).addIndex(index)).addType(type).build();
        SearchResult result = jestClient.execute(search);
        this.logger.info("search data count: " + result.getTotal());
        return result;
    }

    /**
     * 执行sql
     *
     * @param sql
     * @return
     */
    public List<Map<String, Object>> executeSQL (String sql) throws Exception {
        List<Map<String, Object>> returnModels = new ArrayList<>();
        SQLExprParser parser = new ElasticSqlExprParser(sql);
        SQLExpr expr = parser.expr();
        SQLQueryExpr queryExpr = (SQLQueryExpr) expr;
        /*if (parser.getLexer().token() != Token.EOF) {
            throw new ParserException("illegal sql expr : " + sql);
        }*/
        Select select = new SqlParser().parseSelect(queryExpr);
        //通过抽象语法树，封装成自定义的Select，包含了select、from、where group、limit等
        AggregationQueryAction action = null;
        DefaultQueryAction queryAction = null;
        SqlElasticSearchRequestBuilder requestBuilder = null;
        if (select.isAgg) {
            //包含计算的的排序分组的
            action = new AggregationQueryAction(elasticSearchPool.getClient(), select);
            requestBuilder = action.explain();
        } else {
            //封装成自己的Select对象
            Client client = elasticSearchPool.getClient();
            queryAction = new DefaultQueryAction(client, select);
            requestBuilder = queryAction.explain();
        }
        SearchResponse response = (SearchResponse) requestBuilder.get();
        Object queryResult = null;
        if (sql.toUpperCase().indexOf("GROUP") != -1 || sql.toUpperCase().indexOf("SUM") != -1|| sql.toUpperCase().indexOf("COUNT(") != -1) {
            queryResult = response.getAggregations();
        } else {
            queryResult = response.getHits();
        }
        ObjectResult temp = new ObjectResultsExtractor(true, true, true).extractResults(queryResult, true);
        List<String> heads = temp.getHeaders();
        temp.getLines().stream().forEach(one -> {
            Map<String, Object> oneMap = new HashMap<String, Object>();
            for (int i = 0; i < one.size(); i++) {
                Object value = one.get(i);
                /*if(value instanceof Double){
                    Double valueTemp =Double.valueOf((Double)value);
                    DecimalFormat df = new DecimalFormat("######0");
                    value=df.format(valueTemp);
                }*/
                String key = heads.get(i);
                oneMap.put(key, value);
            }
            returnModels.add(oneMap);
        });
        return returnModels;
    }

    /**
     * 执行sql
     * @param sql
     * @return
     */
    public Integer executeCountSQL(String sql) throws Exception {
        SQLExprParser parser = new ElasticSqlExprParser(sql);
        SQLExpr expr = parser.expr();
        SQLQueryExpr queryExpr = (SQLQueryExpr) expr;
        /*if (parser.getLexer().token() != Token.EOF) {
            throw new ParserException("illegal sql expr : " + sql);
        }*/
        Select select = new SqlParser().parseSelect(queryExpr);

        //通过抽象语法树，封装成自定义的Select，包含了select、from、where group、limit等
        SqlElasticSearchRequestBuilder requestBuilder;
        if (select.isAgg) {
            //包含计算的的排序分组的
            AggregationQueryAction action = new AggregationQueryAction(elasticSearchPool.getClient(), select);
            requestBuilder = action.explain();
        } else {
            //封装成自己的Select对象
            Client client = elasticSearchPool.getClient();
            DefaultQueryAction queryAction = new DefaultQueryAction(client, select);
            requestBuilder = queryAction.explain();
        }
        SearchResponse response = (SearchResponse) requestBuilder.get();
        Object queryResult;
        if (sql.toUpperCase().indexOf("GROUP") != -1 || sql.toUpperCase().indexOf("SUM") != -1|| sql.toUpperCase().indexOf("COUNT(") != -1) {
            queryResult = response.getAggregations();
        } else {
            queryResult = response.getHits();
        }
        ObjectResult temp = new ObjectResultsExtractor(true, true, true).extractResults(queryResult, true);
        for (int j = 0; j < temp.getLines().size(); j ++){
            List<Object> one = temp.getLines().get(j);
            for (int i = 0; i < one.size(); i++) {
                Object value = one.get(i);
                if (value instanceof Double){
                    Double valueTemp =Double.valueOf((Double)value);
                    DecimalFormat df = new DecimalFormat("######0");
                    return Integer.parseInt(df.format(valueTemp));
                }
            }
        }
        return 0;
    }

}
